import os
import pathlib
import json
from flask import Flask, g, jsonify, request
from flask_sqlalchemy import SQLAlchemy

from ..config import config

db = SQLAlchemy()


def create_app(config_name):
    """
    Create a Flask application using the app factory pattern.

    :param config_name: Setup app configuration.
    :return: Flask app
    """
    if not os.environ.get('SECRET_KEY'):
        raise NameError('SECRET_KEY must be defined!')  # pragma: no cover

    if not os.environ.get('PG_PASSWORD'):
        raise NameError('PG_PASSWORD must be defined!')  # pragma: no cover

    app = Flask(__name__, instance_relative_config=True)

    app.config.from_object(config[config_name])
    config[config_name].init_app(app)
    app.config.from_pyfile('local_settings.py', silent=True)

    db.init_app(app)

    @app.before_request
    def current_user():
        payload = request.headers.get('Current-User')
        g.current_user = json.loads(payload) if payload else None

    @app.after_request
    def add_header(response):
        server_id = pathlib.Path(os.getcwd()).parts[-1]
        response.headers['Server-Id'] = server_id

        return response

    @app.route('/')
    def index():
        server_id = pathlib.Path(os.getcwd()).parts[-1]
        response = jsonify({'message': f'{server_id} service is up!'})

        return response

    return app



from typing import Generic, List, Optional, Tuple, TypeVar
from enum import Enum
from datetime import datetime
from base64 import b64encode, b64decode
import strawberry
from flask_sqlalchemy import Pagination

T = TypeVar('T')


@strawberry.interface
class Node:
    id: strawberry.ID


@strawberry.type(description='A list of edges.')
class Edge(Generic[T]):
    node: T
    cursor: str


@strawberry.type(description='Information to assist with pagination.')
class PageInfo:
    start_cursor: Optional[str]
    end_cursor: Optional[str]
    has_next_page: bool
    has_previous_page: bool


@strawberry.type(description='A list of edges with pagination information.')
class Connection(Generic[T]):
    edges: List[Edge[T]]
    page_info: PageInfo
    total_count: int
    filtered_count: int
    page_count: int
    current_page: int

    @classmethod
    def load(cls, data: Pagination, counts: Tuple[int, int]):
        return Connection(
            [Edge(item, to_cursor_hash(item.created_at)) for item in data.items],
            PageInfo(
                to_cursor_hash(data.items[0].created_at) if data.items else None,
                to_cursor_hash(data.items[-1].created_at) if data.items else None,
                data.has_next,
                data.has_prev,
            ),
            *counts,  # total_count and filtered_count
            data.pages,  # page_count
            data.page,  # current_page
        )


def to_cursor_hash(created_at: datetime) -> str:
    return str(b64encode(str(created_at).encode('utf-8')), 'utf-8')


def from_cursor_hash(cursor: str) -> datetime:
    return datetime.fromisoformat(str(b64decode(cursor), 'utf-8'))


@strawberry.enum(description='Selection of organisms.')
class OrganismSelect(Enum):
    NONE = ''
    HUMAN = 'human'
    MOUSE = 'mouse'
    RAT = 'rat'


@strawberry.input
class InputWithOrganism:
    organism: Optional[OrganismSelect] = ''



import os


class BaseConfig:
    """Base configuration."""
    DEBUG = False
    TESTING = False
    SECRET_KEY = os.environ.get('SECRET_KEY')
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    # Celery configuration
    CELERY_BROKER_URL = f'redis://{os.environ.get("REDIS_HOST")}:{os.environ.get("REDIS_PORT", 6379)}/0'
    CELERY_RESULT_BACKEND = f'redis://{os.environ.get("REDIS_HOST")}:{os.environ.get("REDIS_PORT", 6379)}/0'
    CELERY_ACCEPT_CONTENT = ['application/json']
    CELERY_TASK_SERIALIZER = 'json'
    CELERY_RESULT_SERIALIZER = 'json'
    CELERY_REDIS_MAX_CONNECTIONS = 5

    @staticmethod
    def init_app(app):
        pass


class DevelopmentConfig(BaseConfig):
    """Development configuration."""
    DEBUG = True
    SQLALCHEMY_DATABASE_URI = \
        f"postgresql://{os.environ.get('PG_USER', 'postgres')}:{os.environ.get('PG_PASSWORD')}" \
        f"@{os.environ.get('PG_HOST', 'localhost')}:{os.environ.get('PG_PORT', 5432)}/{os.environ.get('PG_DB')}"


class TestingConfig(BaseConfig):
    """Testing configuration."""
    TESTING = True
    SQLALCHEMY_DATABASE_URI = \
        f"postgresql://{os.environ.get('PG_USER', 'postgres')}:{os.environ.get('PG_PASSWORD')}" \
        f"@{os.environ.get('PG_HOST', 'localhost')}:{os.environ.get('PG_PORT', 5432)}/" \
        f"{os.environ.get('TEST_DB') or os.environ.get('PG_DB')}"


class ProductionConfig(BaseConfig):
    """Production configuration."""
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = \
        f"postgresql://{os.environ.get('PG_USER', 'postgres')}:{os.environ.get('PG_PASSWORD')}" \
        f"@{os.environ.get('PG_HOST', 'localhost')}:{os.environ.get('PG_PORT', 5432)}/{os.environ.get('PG_DB')}"


config = {
    'development': DevelopmentConfig,
    'testing': TestingConfig,
    'production': ProductionConfig,
    'default': DevelopmentConfig,
}



from datetime import datetime
from typing import Dict, Iterator, List, Optional, Union
from dataclasses import dataclass
import strawberry
from flask_sqlalchemy import BaseQuery
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.schema import Table

from ..app import db


@dataclass(init=False)
class CreateUpdateFields:
    # Keep track when records are created and updated.
    created_at: Optional[datetime] = db.Column(db.DateTime(), index=True, default=datetime.utcnow)
    updated_at: Optional[datetime] = db.Column(db.DateTime(), index=True, default=datetime.utcnow,
                                               onupdate=datetime.utcnow)
    created_by: Optional[int] = db.Column(db.Integer, default=1)
    updated_by: Optional[int] = db.Column(db.Integer)


class ResourceMixin(CreateUpdateFields):
    __table__: Table
    query: BaseQuery

    @classmethod
    def get_by_id(cls, id: Union[int, str, List[str]]):
        try:
            return cls.query.get(id)
        except ValueError:
            return None

    @classmethod
    def get_all(cls):
        return cls.query.all()

    def save(self):
        """
        Save a model instance.

        :return: Model instance
        """
        db.session.add(self)
        db.session.commit()

        return self

    def update(self, update_obj):
        """
        Update a model instance.

        :param update_obj: Model object containing updates
        :return: Model instance
        """
        updates = dict(x for x in update_obj.__dict__.items() if x[0] != '_sa_instance_state')
        for key, value in updates.items():
            setattr(self, key, value)
        self.save()

        return self

    def delete(self) -> None:
        """
        Delete a model instance.

        :return: db.session.commit()'s result
        """
        db.session.delete(self)
        return db.session.commit()

    @classmethod
    def _bulk_insert(cls, data, label: str, dtype: str = '') -> None:
        """
        Bulk insert data to the model and log it. This is much more efficient than adding 1 row at a time in a loop.

        :param data: Data to be saved
        :type data: list
        :param dtype: Data type
        :type dtype: str
        :param label: Label for the output
        :type label: str
        :return: None
        """
        db.engine.execute(cls.__table__.insert(), data)
        print(f'Finished inserting {len(data):,} {(dtype + " ") if dtype else ""}{label}.')

        return None


@dataclass(init=False)
class RefSeqFields:
    id: strawberry.ID = db.Column(db.String(24), primary_key=True)  # RefSeq acc version, e.g. NM_003331.5 / NP_003322.3
    acc: str = db.Column(db.String(16), unique=True, index=True, nullable=False)


class RefseqMixin(RefSeqFields):
    query: BaseQuery

    @classmethod
    def find_by_refseq_id(cls, refseq_id: str):
        """Find a model by its RefSeq accession ID."""
        return cls.query.filter((cls.id == refseq_id) | (cls.acc == refseq_id)).first()


class ExternalResourceMixin:
    __table__: Table
    query: BaseQuery

    # Keep track when records are created and updated.
    created_at = db.Column(db.DateTime(), index=True, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime(), index=True, default=datetime.utcnow, onupdate=datetime.utcnow)
    created_by = db.Column(db.Integer, default=1)
    updated_by = db.Column(db.Integer)

    @classmethod
    def get_by_id(cls, id: Union[int, str, List[str]]):
        try:
            return cls.query.get(id)
        except ValueError:
            return None


def sort_query(model: db.Model, query: BaseQuery, sort_keys: Dict[str, InstrumentedAttribute],
               order_by: Iterator[str]) -> BaseQuery:
    """Sort list with order_by fields, append id_ASC/id_DESC if not present."""
    sort_list = [order.split('_') for order in order_by]
    query = query.order_by(*[sort_keys[sort_key].desc() if sort_order == 'DESC' else sort_keys[sort_key]
                             for (sort_key, sort_order) in sort_list if sort_key in sort_keys])
    if not ('id_ASC' in order_by or 'id_DESC' in order_by):
        query = query.order_by(model.id.desc() if sort_list[0][1] == 'DESC' else model.id)

    return query



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='sea-common',
    version='1.0.1',
    author='SEA Biopharma',
    author_email='sea.biopharma@gmail.com',
    description='SEA Web Services common python packages',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sea-biopharma/sea-common',
    packages=setuptools.find_packages(),
    install_requires=[
        'flask>=1.1.2',
        'flask-sqlalchemy>=2.4.4',
        'strawberry-graphql>=0.45.3',
    ],
    python_requires='>=3.8',
)



from typing import Optional, TypedDict
from flask import g
from strawberry.flask.views import GraphQLView


class UserPayload(TypedDict):
    id: int
    username: str
    email: str


class MyContext(TypedDict):
    current_user: Optional[UserPayload]


class MyGraphQLView(GraphQLView):
    def get_context(self) -> MyContext:
        payload: Optional[UserPayload] = g.current_user

        return {'current_user': payload}



from .app import create_app, db
from .lib.types import Node, Connection, InputWithOrganism
from .lib.util_sqlalchemy import ResourceMixin, RefseqMixin, ExternalResourceMixin, sort_query
from .lib.util_strawberry import MyGraphQLView
